from typing import Set, List, Dict, Tuple, Optional
from enum import Enum
from copy import deepcopy


# TODO Mypy typecheck

class Operation(Enum):
    Union = 1
    Intersection = 2


class Terminal:
    def __init__(self, name: str):  # str = "a"):
        self.name = name

    def __eq__(self, object):
        if isinstance(object, Terminal):
            return object.name == self.name
        return False

    def __hash__(self):
        return hash(self.name)


class Nonterminal:
    def __init__(self, name: str):
        self.name = name


class State:
    def __init__(self, name: str):
        self.name = name


class REG:
    Rules = Dict[Nonterminal, Set[Tuple[Terminal, Optional[Nonterminal]]]]

    def __init__(self, nonterminals: Set[Nonterminal], \
                 terminals: Set[Terminal], \
                 rules: Rules, \
                 init: Nonterminal):
        self.nonterminals = nonterminals
        self.terminals = terminals
        self.rules = rules
        self.init = init


class DFA:
    Transition = Dict[Tuple[State, Terminal], State]

    def __init__(self, states: Set[State], \
                 terminals: Set[Terminal], \
                 transition: Transition, \
                 init: State, \
                 final: Set[State]):
        self.states = states
        self.terminals = terminals
        self.transition = transition
        self.init = init
        self.final = final

    # def total(self) -> DFA:
    def total(self):
        dfa = deepcopy(self)
        #total: Transition
        total = {}

        # better way?
        id = 0
        hell = State(str(id))
        while hell in dfa.states:
            id += 1
            hell.name = str(id)
        dfa.states.add(hell)
        print(dfa.states)

        for state in dfa.states:
            for terminal in dfa.terminals:
                if (state, terminal) in dfa.transition:
                    total[state, terminal] = dfa.transition[state, terminal]
                else:
                    total[state, terminal] = hell

        dfa.transition = total
        return dfa

    # def complement(self) -> DFA:
    def complement(self):
        dfa = deepcopy(self)
        new_final: Set[State]
        for state in dfa.states:
            if state not in dfa.final:
                new_final.add(state)

        dfa.final = new_final
        return dfa

    @staticmethod
    # def composition(dfa_1: DFA, dfa_2: DFA, operation: Operation) -> DFA:
    def composition(dfa_1, dfa_2, operation: Operation):
        dfa = DFA()

        # new states
        for state_1 in dfa_1.states:
            for state_2 in dfa_2.states:
                state = State(state_1.name + "_" + state_2.name)
                dfa.states.add(state)

                # TODO not sure with not dfa.init
                # new init
                if not dfa.init and state_1 == dfa_1.init \
                        and state_2 == dfa_2.init:
                    dfa.init = state

                # new final
                if state_1 in dfa_1.final or state_2 in dfa_2.final:
                    if operation == Operation.Union:
                        dfa.final.add(state)
                    elif operation == Operation.Intersection and \
                            state_1 in dfa_1.final and \
                            state_2 in dfa_2.final:
                        dfa.final.add(state)

        # new terminals
        dfa.terminals.add(dfa_1.terminals)
        dfa.terminals.add(dfa_2.terminals)
        # ATTENTION: terminals may differ -> make total at the end?

        # TODO new transition

        return dfa

    @staticmethod
    # def union(dfa_1: DFA, dfa_2: DFA) -> DFA:
    def union(dfa_1, dfa_2):
        composition(dfa_1, dfa_2, Union)

    @staticmethod
    # def intersection(dfa_1: DFA, dfa_2: DFA) -> DFA:
    def intersection(dfa_1, dfa_2):
        composition(dfa_1, dfa_2, Intersection)


class NFA:
    def __init__(self, states: Set[State], \
                 terminals: Set[Terminal], \
                 transition: Dict[Tuple[State, Terminal], Set[State]], \
                 init: State, \
                 final: Set[State]):
        self.states = states
        self.terminals = terminal
        self.transition = transition
        self.init = init
        self.final = final

